use crate::constants::random_useragent;
use crate::Client;
use isahc::http::HeaderMap;
use isahc::ReadResponseExt;
#[allow(unused_imports)]
use log::debug;

/// this is what we look for to find the bearer token
const SEARCH_STRING: &str = "AAAAAAAAA";
/// this is what we look for to get the guest_id string, it has a `=` at the end because we don't
/// wanna get something like guest_id_marketing accidentally
/// (don't mind the fact that guest_id_marketing currently holds the same value as guest_id)
const GUEST_ID_SEARCH_STRING: &str = "guest_id=";
/// this is what we look for to find the guest token ("gt")
const GUEST_TOKEN_SEARCH_STRING: &str = "\"gt=";
/// this is what we look for to find the main.*.js file
const MAIN_JS_SEARCH_STRING: &str = "main.";

#[derive(Clone, Debug)]
pub enum BearerError {
    CouldntGetTwitterDotCom(isahc::Error),
    CouldntFindMainJsUrl,
    CouldntGetMainDotJs(isahc::Error),
    CouldntGetGuestIdCookie,
    CouldntFindBearerToken,
    CouldntGetGuestToken,
}

/// `let (bearer_token, guest_id, guest_token) = get_bearer_and_guest_stuff(...);`
pub fn get_bearer_and_guest_stuff(
    client: &Client,
) -> Result<(String, String, String), BearerError> {
    let url = "https://twitter.com";
    let mut headers = HeaderMap::new();
    let user_agent = random_useragent();
    headers.insert("user-agent", user_agent.parse().unwrap());

    let mut response = client
        .get(url, &headers)
        .map_err(BearerError::CouldntGetTwitterDotCom)?;
    let body_headers = response.headers();
    // get the set-cookie header for the guest id
    let guest_id_header = body_headers
        .get_all("set-cookie")
        .iter()
        .filter(|v| {
            v.to_str()
                .unwrap_or_default()
                .starts_with(GUEST_ID_SEARCH_STRING)
        })
        .next()
        .ok_or(BearerError::CouldntGetGuestIdCookie)?;
    let guest_id_cookie = guest_id_header
        .to_str()
        .map_err(|_| BearerError::CouldntGetGuestIdCookie)?;
    let guest_id = guest_id_cookie[0..guest_id_cookie.find(';').unwrap_or(guest_id_cookie.len())]
        .split_once('=')
        .ok_or(BearerError::CouldntGetGuestIdCookie)?
        .1
        .to_string();
    let body = response.text().unwrap();
    let mut mainjs_url = None;
    let mut guest_token = None;
    // fixme! this is really fucking cursed
    for line in body.lines() {
        if line.contains(MAIN_JS_SEARCH_STRING) {
            let main_js_start = line.find(MAIN_JS_SEARCH_STRING).unwrap();
            let main_js_end = line[main_js_start..].find('"').unwrap() + main_js_start;
            let main_js_url_start = line[0..main_js_start].rfind('"').unwrap();
            let main_js_url = &line[main_js_url_start + 1..main_js_end];
            mainjs_url = Some(main_js_url.to_string());
        }
    }
    for line in body.lines() {
        if line.contains(GUEST_TOKEN_SEARCH_STRING) {
            let guest_start = line.find(GUEST_TOKEN_SEARCH_STRING).unwrap();
            let guest_end = line[guest_start..].find(';').unwrap() + guest_start;
            let gt = &line[guest_start + 1..guest_end];
            guest_token = Some(gt.to_string());
        }
    }
    let mainjs_url = mainjs_url.ok_or(BearerError::CouldntFindMainJsUrl)?;
    let guest_token = guest_token
        .ok_or(BearerError::CouldntGetGuestToken)?
        .split_once('=')
        .ok_or(BearerError::CouldntGetGuestToken)?
        .1
        .to_string();
    // now get the main.js file
    let mut response = client
        .get(&mainjs_url, &headers)
        .map_err(BearerError::CouldntGetMainDotJs)?;
    let body = response.text().unwrap();
    let mut token = None;
    for line in body.lines() {
        if line.contains(SEARCH_STRING) {
            let bearer_token_start = line.find(SEARCH_STRING).unwrap();
            let bearer_token_end =
                line[bearer_token_start..].find('"').unwrap() + bearer_token_start;
            let bearer_token = &line[bearer_token_start..bearer_token_end];
            token = Some(bearer_token.to_string());
        }
    }

    let token = token.ok_or(BearerError::CouldntFindBearerToken)?;
    Ok((token, guest_id, guest_token))
}
